package org.saager.enumerator;

import static org.saager.enumerator.Predicate.TAUTOLOGIC;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Enumerator
 * 
 * The Enumerator enable the consistent use of the enhanced for-loop of Java 5 SE
 * 
 * Any object eligible for an enhanced for-loop can be wrapped by an Enumerator to enable 
 * <ol>
 * <li>removal</li>
 * <li>indexing</li>
 * <li>introspection</li>
 * </ol>
 * 
 * In addition the Enumerator can be filtered using Predicates and appended by other Enumerators
 * 
 * @see Iterator
 * @see Predicate
 * @see Cursor
 * 
 * @author Carsten Saager
 * 
 * @param <T>
 */
public class Enumerator<T> implements Iterable<Cursor<T>> {
	/**
	 * append one or more Enumerators to this.
	 * 
	 * @param tail
	 * @return new iterator that will chain all in the given order, this first
	 */
	public Enumerator<T> append( final Enumerator<? extends T> ... tail) {
		return new Enumerator<T>(new Iterable<T>(){
			
			public Iterator<T> iterator() {
				
				return new LiveIterator<T>(){
					int nextIter;
					Iterator<? extends T> currIt=iterable.iterator();
					Iterator<? extends T> nextIt=iterable.iterator();
					public boolean hasNext() {
						if (currIt.hasNext()) return true;
						return advance();
					}
					private boolean advance() {
						while (nextIter<tail.length) {
							nextIt = tail[nextIter++].iterable.iterator();
							if (nextIt.hasNext())  return true;
						}
						nextIt=null;
						return false;
					}
					public T next() {
						if (currIt.hasNext()) return currIt.next();
						else if (nextIt!=null || advance()) {
							currIt=nextIt;
							nextIt=null;
						} else {
							throw new NoSuchElementException();
						}
						return currIt.next();
					}

					public void remove() {
						currIt.remove();
					}
					@SuppressWarnings("unchecked")
					public void set(T obj) {
						if (currIt instanceof LiveIterator) {
							LiveIterator<T> modifier = (LiveIterator<T>) currIt;
							modifier.set(obj);
						} else {
							throw new IllegalStateException();
						}
					}
					
				};
			}});
	}
	
	private final Iterable<T> iterable;
	private final Predicate<? super T> filter;

	/**
	 * An unfiltered Enumerator
	 * @param iterable
	 */
	public Enumerator(Iterable<? extends T> iterable) {
		this(iterable,Predicate.TAUTOLOGIC);
	}

	/**
	 * Iterating with preset filter
	 * @param iterable
	 * @param filter
	 */
	@SuppressWarnings("unchecked")
	public Enumerator(Iterable<? extends T> iterable, Predicate<? extends Object> filter) {
		this.iterable = (Iterable<T>) iterable;
		this.filter = (Predicate<? super T>) filter;
	}
	/**
	 * Restricts results to object matching this Predicate
	 * 
	 * @param predicate
	 * @return
	 */
	public Enumerator<T> only(Predicate<? super T> predicate) {
		return new Enumerator<T>(iterable, filter==TAUTOLOGIC?predicate:filter.and(predicate));
	}
	/**
	 * Also returns results from this Predicate
	 * @param predicate
	 * @return
	 */
	public Enumerator<T> also(Predicate<? super T> predicate) {
		return new Enumerator<T>(iterable, filter==TAUTOLOGIC?predicate:filter.or(predicate));
	}
	/**
	 * XOR on Predicates
	 * @param predicate
	 * @return
	 */
	public Enumerator<T> either(Predicate<? super T> predicate) {
		return new Enumerator<T>(iterable, filter==TAUTOLOGIC?predicate:filter.either(predicate));
	}
	/**
	 * Complement of predicated set
	 * @return
	 */
	public Enumerator<T> invert() {
		return new Enumerator<T>(iterable, filter.not());

	}
	/**
	 * Create an Enumerator that returns only element of this sub-class.
	 * 
	 * This is not implemented via Predicates, because too dangerous to loose type safety,
	 * thus all Predicate function can be usedwithout any restriction.
	 * 
	 * @param <S>
	 * @param clazz
	 * @return
	 */
	public <S extends T> Enumerator<S> specialize(final Class<S> clazz){
		return new Enumerator<S>(new Iterable<S>() {
			
			public Iterator<S> iterator() {
				return new LiveIterator<S>() {
					Iterator<T >delegate = iterable.iterator();
					S next;
					private boolean blockRemove;
					public boolean hasNext() {
						blockRemove=true;
						if (next==null){
							while (delegate.hasNext()) {
								T cand = delegate.next();
								if (clazz.isInstance(cand)) {
									next = clazz.cast(cand);
									break;
								}
							}
						}
						return next!=null;
					}
					
					public S next() {
						S tmp = next;
						if (next==null) {
							hasNext();
							if (next==null) throw new NoSuchElementException();
						} else next=null;
						blockRemove = false;
						return tmp;
					}

					public void remove() {
						if (blockRemove) throw new IllegalStateException();
						blockRemove = true;
						delegate.remove();						
					}

					@SuppressWarnings("unchecked")
					public void set(S obj) {
						if (delegate instanceof LiveIterator) {
							LiveIterator<S> modifier = (LiveIterator<S>) delegate;
							modifier.set(obj);
							
						}
					}
					
				};
			}});
	}
	private final class IndexedIterator implements Iterator<Cursor<T>> {
		Iterator<T> inner = iterable.iterator();

		private Cursor<T> next;
		private Cursor<T> removable;

		private int index;
		public boolean hasNext() {
			if (next==null){
				removable=null;
				while (inner.hasNext()) {
					T cand = inner.next();
					if (filter.matches(cand)) {
						next = new Cursor<T>(cand, index++) {

							@Override
							public boolean remove() {
								if (removable==this)
									try {
										inner.remove();
										removable=null;
										return true;
									} catch (UnsupportedOperationException usoe) {
										return false;
									}
								else {
									throw new IllegalStateException();
								}
							}

							@Override
							public boolean hasSuccessor() {
								return (index+1 < IndexedIterator.this.index) ? true : hasNext();
							}

							@Override
							public boolean update(T obj) {
								if (inner instanceof LiveIterator) {
									if (removable==this) {
										LiveIterator<T> modifier = (LiveIterator<T>) inner;
										modifier.set(obj);
										return true;
									} else {
										throw new IllegalStateException();
									}
								}
								return false;
							}

						};
						break;
					}
				}
			}
			return next!=null;
		}
		public Cursor<T> next() {
			if (hasNext()) {
				removable=next;
				next = null;
				return removable;
			} else throw new NoSuchElementException();
		}

		public void remove() {
			removable.remove();
		}

	}
	/* (non-Javadoc)
	 * @see java.lang.Iterable#iterator()
	 */
	public Iterator<Cursor<T>> iterator() {
		return new IndexedIterator();
	}
	
}
